//==============================================
//  EDID Emulator
//  EDID_Emulator.ino
//
//  2023/05/15 Ver 0.0
//  2024/05/25 Ver 0.1
//
//  Copyright (C)2023-2024 FREE WING, Y.Sakamoto
//
//  http://www.neko.ne.jp/~freewing/
//==============================================

#include <Wire.h>

// I2C SCL freq.
#define I2C_STANDARD_MODE    100000UL /** 100kbps */
#define I2C_FREQ I2C_STANDARD_MODE

// EDID I2C Address = 0x50 (same as 24LC01/ 24LC02)
#define EDID_I2C_ADDRESS 0x50
#define EDID_LENGTH 128
#define EDID_REQUEST_NONE 255

// Macro
// char toHexStr_tmp[5];
// #define toHexStr(b) utoa((b), toHexStr_tmp, HEX)
#define toHexChar(b) (b) + '0' + ((b) >= 10) * ('A' - '0' - 10)

char* toHexStr(byte n, bool padding_zero = true)
{
  static char buf[sizeof(byte) * 2 + 1]; // Assumes 8-bit chars plus zero byte.
  char *str = &buf[sizeof(buf) - 1];

  *str = '\0';

  do {
    char c = n & 0x0F;
    n >>= 4;

    *--str = c + '0' + (c >= 10) * ('A' - '0' - 10);
  } while((padding_zero & str != buf) || n);

  return str;
}

// I2C Slave function
void onReceive(int len);
void onRequest();
void onRequestExtra();

#if defined(ARDUINO_ARCH_STM32)
  #include "arch_stm32.h"
#elif defined(__LGT8F__) || defined(__LGT8FX8P__)
  #include "arch_lgt8.h"
#elif defined(ARDUINO_ARCH_AVR)
  #include "arch_avr.h"
#elif defined(ARDUINO_ARCH_ESP32)
  #include "arch_esp32.h"
#elif defined(ARDUINO_ARCH_RP2040)
  #include "arch_rp2040.h"
#elif defined(ARDUINO_ARCH_RENESAS)
  // Arduino UNO R4
  #include "arch_uno_r4.h"
#else
  #error "Error: Not Support this Board"
#endif

#if defined (NOT_USE_SERIAL)
  #include "NullSerial.h"
  #define SerialDebug   NullSerial
#else
  #if defined(SERIAL_PORT_MONITOR)
    #define SerialDebug   SERIAL_PORT_MONITOR
  #else
    #define SerialDebug   Serial
  #endif
#endif

// I2C OLED SSD1306 64x32 0.49 inch
#define OLED_I2C_ADDRESS 0x3C
#define OLED_BRIGHTNESS 24
#include "oled.h"

// EDID define table
#include "edid.h"

// enum
enum I2C_STATUS {
  // I2C Master: NONE -> WRITE -> REQ. -> READ -> DONE
  // I2C Slave : NONE -> RECV -> REQ. -> DONE
  NONE,
  WRITE,
  RECV,
  REQ,
  READ,
  DONE
};

// enum to String
const char I2C_STATUS_NONE[]  PROGMEM = "None";
const char I2C_STATUS_WRITE[] PROGMEM = "Writ";
const char I2C_STATUS_RECV[]  PROGMEM = "Recv";
const char I2C_STATUS_REQ[]   PROGMEM = "Req.";
const char I2C_STATUS_READ[]  PROGMEM = "Read";
const char I2C_STATUS_DONE[]  PROGMEM = "Done";

const char* const I2C_STATUS_NAME[] PROGMEM = {
  I2C_STATUS_NONE,
  I2C_STATUS_WRITE,
  I2C_STATUS_RECV,
  I2C_STATUS_REQ,
  I2C_STATUS_READ,
  I2C_STATUS_DONE
};

/** Current Selected EDID table index */
volatile uint8_t edidIndex = 0;

/** I2C status Sink device */
volatile I2C_STATUS i2cMasterStatus = NONE;
/** I2C Error status */
volatile bool isI2cMasterError = false;

/** I2C status Source device */
volatile I2C_STATUS i2cSlaveStatus = NONE;
/** I2C Error status */
volatile bool isI2cSlaveError = false;

/** Requested EDID memory address */
volatile uint8_t requestMemoryAddress = EDID_REQUEST_NONE;
/** EDID information from Display device (in AVR Used as work for EDID transfer) */
byte displayEdidInfo[EDID_LENGTH * 2];

// Get EDID info. from Display device
bool readI2cMemory(byte readMemoryAddress, byte* displayEdidInfo, const byte readLength = EDID_LENGTH) {
  SerialDebug.print(F("readI2cMemory: "));
  SerialDebug.println(toHexStr(readMemoryAddress));

  i2cMasterStatus = WRITE;

  int retry = 3;
  while (--retry > 0) {
    WireMaster.beginTransmission(EDID_I2C_ADDRESS);
    WireMaster.write(readMemoryAddress);
    if (WireMaster.endTransmission() == 0) {
      // 0 = success
      break;
    }

    SerialDebug.print(F("retry: "));
    SerialDebug.println(toHexStr(retry, false));

    delay(1);
  }
  delay(1);

  // I2C Error
  if (retry == 0) {
    return false;
  }

  i2cMasterStatus = REQ;

  // uint8_t TwoWire::requestFrom(uint8_t, uint8_t)
  if (WireMaster.requestFrom((uint8_t)EDID_I2C_ADDRESS, (uint8_t)readLength) != readLength) {
    // I2C Error
    return false;
  }

  i2cMasterStatus = READ;

  int i = 0x00;
  for (; i<readLength; ++i) {
    int r = WireMaster.read();
    if (r == -1) {
      // I2C Error
      break;
    }
    displayEdidInfo[i] = (byte)r;

    // SerialDebug.print((byte)r, HEX);
    SerialDebug.print(toHexStr((byte)r));
    SerialDebug.write(' ');
    if ((i % 16) == 15) SerialDebug.println();
  }
  SerialDebug.println();

  delay(1);

  return (i == readLength);
}

bool readEdidInfoFromDisplay() {
  // 0x00 - 0x7F
  if (!readI2cMemory(0x00, &displayEdidInfo[0x00])) return false;

  // has EDID Extension ?
  if (displayEdidInfo[0x7E] != 0x00) {
    // Support one Extension block
    // 0x80 - 0xFF
    if (!readI2cMemory(0x80, &displayEdidInfo[0x80])) return false;
  }

  i2cMasterStatus = DONE;
  return true;
}

// I2C Slave function
void onReceive(int len) {
  i2cSlaveStatus = RECV;

  int r;
  for (int i=0; i<len; ++i) {
    r = WireSlave.read();
    if (r == -1) {
      isI2cSlaveError = true;
      return;
    }
  }

  requestMemoryAddress = (byte)r;
}

// I2C Slave function
void onRequest() {
  i2cSlaveStatus = REQ;

  size_t sendSize = 1;
  // macBook Intel
  if (requestMemoryAddress == 0x08) sendSize = 2;
  // Switch GAME machine
  if ((requestMemoryAddress & 0x0f) == 0x00) sendSize = 0x10;
  // 0x00 or 0x80 = EDID_LENGTH
  if ((requestMemoryAddress & 0x7f) == 0x00) sendSize = EDID_LENGTH;

  // EDID Table or Display Raw EDID info
#if defined(IS_AVR)
  if (edidIndex < EDID_DISPLAY_RAW_INDEX) {
    memcpy_P(displayEdidInfo, edid_mem_table[edidIndex], EDID_LENGTH*2);
  }
  const uint8_t* ptr = &displayEdidInfo[requestMemoryAddress];
#else
  const uint8_t* ptr = (edidIndex < EDID_DISPLAY_RAW_INDEX)?
    &edid_mem_table[edidIndex][requestMemoryAddress] : &displayEdidInfo[requestMemoryAddress];
#endif

  if (WireSlave.write(ptr, sendSize) != sendSize) {
    isI2cSlaveError = true;
    return;
  }

  i2cSlaveStatus = DONE;
}

// Button handler
volatile bool isButtonPushed = false;

#define checkButtonPush() (digitalRead(USER_BUTTON) == BUTTON_ON)

void buttonPushFunc() {
  isButtonPushed = true;
}

void buttonClear() {
  isButtonPushed = false;
}

// HPG Hot Plus Detect handler
volatile bool isDisplayEnable = false;

void displayDissconnectFunc() {
  displayDisabled();
}

void displayDisabled() {
  isDisplayEnable = false;
  disconnectHpd();
}

void displayEnabled() {
  isDisplayEnable = true;
}

#define checkDisplayHpd() (digitalRead(HPD_DISPLAY) == HPD_ENABLE)

// Control PC HPG
void disconnectHpd() {
  digitalWrite(HPD_COMPUTER, HPD_DISABLE);
}

void connectHpd() {
  digitalWrite(HPD_COMPUTER, HPD_ENABLE);
}

// Blink LED
void blinkLed() {
#ifdef USER_LED
  digitalWrite(USER_LED, LED_ON);   delay(5);
  digitalWrite(USER_LED, LED_OFF);  delay(100);
#endif
}

#if defined(IS_AVR)
// for AVR interrupt
ISR(PCINT0_vect)
{
  if (checkButtonPush()) {
    buttonPushFunc();
  }

  if (!checkDisplayHpd()) {
    displayDissconnectFunc();
  }
}
#endif

// Setup
void setup() {
#ifdef USER_LED
  pinMode(USER_LED, OUTPUT);
#endif
  pinMode(USER_BUTTON, INPUT_PULLUP);

  pinMode(HPD_COMPUTER, HPD_COMPUTER_OUTPUT);
  digitalWrite(HPD_COMPUTER, HPD_DISABLE);
  pinMode(HPD_DISPLAY, HPD_DISPLAY_INPUT);

  // Init OLED
  delay(50);
  Oled_begin();
  Oled_clear();
  // Splash Screen 
  Oled_print(0, 0, F("EDID"), false, true);
  Oled_print(0, 2, F(" emu."), true, true);
  Oled_end();

  // Blink LED
  blinkLed();
  blinkLed();
  blinkLed();

  i2cMasterStatus = NONE;
  i2cSlaveStatus = NONE;

  SerialDebug.begin(115200);
  while(!SerialDebug && (millis() < 5000));
  SerialDebug.println(F("setup()"));

  // Wait Splash Screen at least 1sec
  while(millis() < 1000);

  // Initialize
  buttonClear();
  displayDisabled();

  #ifndef NOT_USE_INT
    // Setup Interrupt
    #if defined(IS_AVR)
      // https://thewanderingengineer.com/2014/08/11/arduino-pin-change-interrupts/
      cli();
      PCICR  |= AVR_PCICR;  // Pin Change Interrupt Control Register
      PCMSK0 |= AVR_PCMSK0; // Pin Change Enable Mask
      sei();
    #else
      attachInterrupt(USER_BUTTON, buttonPushFunc, FALLING);
      attachInterrupt(HPD_DISPLAY, displayDissconnectFunc, FALLING);
    #endif
  #endif

  WireSlave_begin();
}

// Loop
void loop() {
  static byte loopCount;

  #ifdef NOT_USE_INT
    if (checkButtonPush()) {
      while (checkButtonPush()) ;
      buttonPushFunc();
    }

    if (isDisplayEnable) {
      if (!checkDisplayHpd()) {
        displayDissconnectFunc();
      }
    }
  #endif

  delay(100);

  // SerialDebug.println(F("loop()"));
  // SerialDebug.printf("adrs: %X %d %d %d %d\r\n", requestMemoryAddress, isI2cMasterError, isI2cSlaveError, edidIndex, isDisplayEnable);
  // SerialDebug.println(toHexStr(requestMemoryAddress));

  if (!isOledWorking) {
    Oled_begin();
  }

  Oled_print(0, 0, toHexChar(edidIndex), true, true);
  if (edidIndex < EDID_DISPLAY_RAW_INDEX) {
#if defined(IS_AVR)
    Oled_print_P(1, 1, &(edid_name_table[edidIndex]));
#else
    Oled_print(1, 1, edid_name_table[edidIndex]);
#endif
  } else {
    Oled_print(1, 1, F("Raw Disp."));
  }

  Oled_print(9, 0, (loopCount++ & 1)? '\'': ' ');
  {
    Oled_print(0, 2, F("PC"), (i2cSlaveStatus == NONE));
    Oled_print(-1, -1, ':');
    Oled_print(-1, -1, toHexStr(requestMemoryAddress));
#if defined(IS_AVR)
    Oled_print_P(6, 2, &(I2C_STATUS_NAME[i2cSlaveStatus]), isI2cSlaveError);
#else
    Oled_print(6, 2, I2C_STATUS_NAME[i2cSlaveStatus], isI2cSlaveError);
#endif
    // Display status
    Oled_print(0, 3, F("DISP"), !isDisplayEnable);
    Oled_print(-1, -1, ':');
#if defined(IS_AVR)
    Oled_print_P(6, 3, &(I2C_STATUS_NAME[i2cMasterStatus]), isI2cMasterError);
#else
    Oled_print(6, 3, I2C_STATUS_NAME[i2cMasterStatus], isI2cMasterError);
#endif
  }

  if (!isDisplayEnable) {
    if (checkDisplayHpd()) {
      displayEnabled();

      delay(1200);

      if (isDisplayEnable && checkDisplayHpd()) {
        // Connect HPD
        connectHpd();
      }
    }
  }

  if (isButtonPushed) {
    blinkLed();

    // 0-n = EDID Table, EDID_DISPLAY_RAW_INDEX = Display Raw EDID info.
    while (true) {
      ++edidIndex;

      // EDID_DISPLAY_RAW_INDEX = Display Raw EDID info.
      if ((edidIndex == edid_table_size) || (edidIndex == EDID_DISPLAY_RAW_INDEX)) {
        edidIndex = EDID_DISPLAY_RAW_INDEX;

        blinkLed();
        blinkLed();

        // Check Display Enable
        if (!isDisplayEnable) {
          continue;
        }

        if (isOledWorking) {
          Oled_end();
        }

        // Read Display Raw EDID info.
        WireMaster_begin();
        bool success = readEdidInfoFromDisplay();
        isI2cMasterError = !success;
        WireMaster_end();

        if (success) break;
      }

      // Loop Index = 0
      if (edidIndex >= EDID_DISPLAY_RAW_INDEX) edidIndex = 0;

      // Check Valid EDID info.
      if (pgm_read_byte(&edid_mem_table[edidIndex][0]) == 0x00 && pgm_read_byte(&edid_mem_table[edidIndex][1]) == 0xFF) break;
    }

    SerialDebug.print(F("edidIndex: "));
    SerialDebug.println(toHexStr(edidIndex, false));

    SerialDebug.print(F("isDisplayEnable: "));
    SerialDebug.println(toHexStr(isDisplayEnable, false));

    // Init
    requestMemoryAddress = EDID_REQUEST_NONE;
    i2cSlaveStatus = NONE;

    if (isDisplayEnable) {
      // Disconnect HPD
      disconnectHpd();

      // HPD delay over 1sec
      delay(1200);

      // Connect HPD
      connectHpd();
    }

    buttonClear();
  }
}
